/*
 * Description: A script
 * Version: 0.1
 * Author: Daniel Andrade
 * Date: 26/01/2017
 * This code may not be reused without proper permission from its creator.
 */

/**
 * @property {number} Id
 * @property {number} RenterId
 * @property {number} GoodId
 * @property {Date} StartDateTime
 * @property {Date} EndDateTime
 * @property {number} PickUpAddressId
 * @property {number} FixedCredits
 * @property {number} VariableCredits
 * @property {number} ExtraCredits
 * @property {string} Observations
 * @property {Object} Meta
 *
 * @property {Date} CheckInTime
 * @property {string} CheckInObservations
 * @property {Date} CheckOutTime
 * @property {string} CheckOutObservations
 *
 * @property {number} Status
 *
 * @property {Good} Good
 * @property {User} Renter
 * @property {Address} PickupAddress
 *
 * @extends ORM.Entity
 * @constructor
 */
function Reservation(){
    this.Parent(null, arguments, Reservation);

    if (!this.PickupAddress) this.PickupAddress = new Address();
    if (!this.Meta){
        this.Meta = {};

        if (this.Good){
            if (this.Good.Type.HasPickupAddress){
                this.Meta.pickupAddress = '';
                this.Meta.pickupPlace = null;
            }
            if (this.Good.Type.HasDestinationAddress){
                this.Meta.destinationAddress = '';
                this.Meta.destinationPlace = null;
            }
            if (this.Good.Type.HasPickupAddress && this.Good.Type.HasDestinationAddress){
                this.Meta.distance = 0;
            }
        }
    }

    if (!this.FixedCredits && this.Good){
        this.FixedCredits = this.Good.DefaultCost;
    }

}
Reservation.inherit(Entity);
Reservation.Attributes = [
    new Entity.Attributes.Integer('Id'),
    new Entity.Attributes.Integer('RenterId'),
    new Entity.Attributes.Integer('GoodId'),
    new Entity.Attributes.UnixTimestamp('StartDateTime'),
    new Entity.Attributes.UnixTimestamp('EndDateTime'),
    new Entity.Attributes.Integer('PickUpAddressId'),
    new Entity.Attributes.Integer('FixedCredits'),
    new Entity.Attributes.Integer('VariableCredits'),
    new Entity.Attributes.Integer('ExtraCredits'),
    new Entity.Attributes.String('Observations'),
    new Entity.Attributes.Json('Meta'),

    new Entity.Attributes.UnixTimestamp('CheckInTime'),
    new Entity.Attributes.String('CheckInObservations'),
    new Entity.Attributes.UnixTimestamp('CheckOutTime'),
    new Entity.Attributes.String('CheckOutObservations'),

    new Entity.Attributes.Integer('Status'),

    new Entity.Attributes.Boolean('ManualPayment'),
    new Entity.Attributes.String('PaymentCode'),
    new Entity.Attributes.Integer('PaymentUserId', true),

    new Entity.Attributes.Entity('Good', false),
    new Entity.Attributes.Entity('Renter', false),
    new Entity.Attributes.Entity('PickupAddress', false)
];
Entity.ClassMap.Register(Reservation);

Reservation.prototype.getCost = function () {
    return this.FixedCredits + this.VariableCredits;
};

Reservation.prototype.calculateVariableCost = function () {
    if (this.Good && this.Good.Type){
        var good = this.Good;
        var type = this.Good.Type;

        if (good.VariableCostMeasurement == Good.UsageMeasurement.Kilometers){
            // Calc route
            if (this.Meta.pickupPlace && this.Meta.destinationPlace){
                var res = this;

                this.calculateRouteDistance(function (distance) {
                    res.Meta.distance = Math.round(distance/10)/100;

                    res.VariableCredits = Math.round(distance/1000) * good.VariableCost;
                });
            }else{
                this.VariableCredits = 0;
            }
        }else if (good.VariableCostMeasurement == Good.UsageMeasurement.Miles){
            // Calc distance
            if (this.Meta.pickupPlace && this.Meta.destinationPlace){
                var km = this.calculateDirectDistance()/1000;
                var miles = km * 0.621371;

                this.Meta.distance = Math.round(km*100)/100;
                this.VariableCredits = Math.round(miles) * good.VariableCost;
            }else{
                this.VariableCredits = 0;
            }
        }else if (good.VariableCostMeasurement == Good.UsageMeasurement.Minutes){
            // Calc distance and divide by average speed
            if (this.Meta.pickupPlace && this.Meta.destinationPlace){

            }else{
                this.VariableCredits = 0;
            }
        }else if (good.VariableCostMeasurement == Good.UsageMeasurement.Hours){
            // Calc period interval
            if (this.StartDateTime && this.EndDateTime){
                this.VariableCredits = this.calculateDateDifference(1000 * 60 * 60) * good.VariableCost;
            }else{
                this.VariableCredits = 0;
            }
        }else if (good.VariableCostMeasurement == Good.UsageMeasurement.Days){
            // Calc period interval
            if (this.StartDateTime && this.EndDateTime){
                this.VariableCredits = this.calculateDateDifference(1000 * 60 * 60 * 24) * good.VariableCost;
            }else{
                this.VariableCredits = 0;
            }
        }

        return this.VariableCredits;
    }
};

Reservation.prototype.getDirectionService = function () {
    if (!Reservation.ds && google && google.maps && google.maps.DirectionsService){
        Reservation.ds = new google.maps.DirectionsService();
    }

    return Reservation.ds;
};

Reservation.prototype.calculateRouteDistance = function (fn) {
    var ds = this.getDirectionService();

    ds.route({
        origin: this.Meta.pickupPlace.geometry.location,
        destination: this.Meta.destinationPlace.geometry.location,
        travelMode: 'DRIVING'
    }, function(response, status) {
        if (status === 'OK') {
            var route = response.routes[0];

            var dist = 0;

            for (var i = 0; i < route.legs.length; i++){
                var leg = route.legs[i];
                dist += leg.distance.value;
            }

            debugger;
            fn(dist);
        }else{
            console.log('Failed to fetch directions');
        }
    });
};
Reservation.prototype.calculateDirectDistance = function () {
    if (google && google.maps && google.maps.geometry){
        return google.maps.geometry.spherical.computeDistanceBetween(
            this.Meta.pickupPlace.geometry.location,
            this.Meta.destinationPlace.geometry.location);
    }
    return 0;
};

Reservation.prototype.calculateDateDifference = function (unit) {
    var diff = this.EndDateTime.getTime() - this.StartDateTime.getTime();

    return diff > 0 ? Math.ceil(diff/unit) : 0;
};

Reservation.prototype.calculateCost = function () {
    return this.FixedCredits + this.calculateVariableCost();
};


Reservation.Status = {
    Pending: 0,
    Approved: 1,
    Started: 2,
    Finished: 3,
    Canceled: 4,
    Denied: 5,

    _strings: {
        0: 'Em análise',
        1: 'Aprovada',
        2: 'Inciada',
        3: 'Finalizada',
        4: 'Cancelada',
        5: 'Negada'
    },
    _classes: {
        0: 'warning',
        1: 'primary',
        2: 'success',
        3: 'info',
        4: 'danger',
        5: 'danger'
    }
};